#include "stdafx.h"
#include <utility>
#include <Windows.h>
#include "handle_watcher.h"

using std::move;
using std::function;


handle_watcher::handle_watcher():
    waitObject_(nullptr),
    handle_(nullptr, &CloseHandle),
    workItem_()
{}

bool handle_watcher::watch_this(void*&& handle, function<void()> workItem)
{
    return watch_this(handle_ptr(handle, &CloseHandle), move(workItem));
}

bool handle_watcher::watch_this(handle_ptr handle, function<void()> workItem)
{
    cleanup();

    if (!handle)
        return false;

    workItem_ = move(workItem);
    handle_ = move(handle);

    HANDLE wo;
    if (RegisterWaitForSingleObject(&wo, handle_.get(), &handle_watcher::dispatch_signaled, this, INFINITE, WT_EXECUTEONLYONCE))
        waitObject_ = wo;

    return true;
}

handle_ptr handle_watcher::cleanup()
{
    if (HANDLE wo = waitObject_.exchange(nullptr))
    {
        // wait for possibly queued callback to complete
        UnregisterWaitEx(wo, INVALID_HANDLE_VALUE);

        if (bool signaled = !move(workItem_))
            // reset postponed until after unregistration
            handle_ = nullptr;
        return move(handle_);
    }

    // note: handle already reset by signaled()
    return handle_ptr(nullptr, &CloseHandle);
}

void handle_watcher::signaled()
{
    // invoke callback before unregistration, such that an interim call to cleanup() blocks
    workItem_();

    // handle and callback remain unused thereafter;
    // resetting is thread-safe in regards to watch_this() and cleanup(), which wait for signaled() to complete
    workItem_ = nullptr;
    handle_ptr handle = move(handle_); // postpone reset until after unregistration

    if (HANDLE wo = waitObject_.exchange(nullptr))
        // unregisters unconditionally (without waiting for already queued callbacks to get called)
        UnregisterWaitEx(wo, nullptr);
    else
        // postpone reset until after unregistration in cleanup()
        handle_ = move(handle);

    // attention: don't access 'this' anymore (main thread destroying the handle watcher might kick in) ...
}
